/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.cache.api;

/**
 * This interface is the functional (applicative) view for the Caching
 *  Service.  All   accesses  to  the  cache  are   made  through  the
 * Cache instance provided to the applicative level.
 *
 *  The primary  purpose  of the  caching  service is  to improve  the
 * overall system performance.  The performance gain is from avoidance
 * of costs due to  I/O operations related to storage/network devices,
 *  recreation-related costs,  etc.  This  service is  to  be used  by
 *  several   services  such  as  the  persistence   service  and  the
 * replication service, among others.
 *
 * @author Luciano Garcia-Banuelos (Luciano.Garcia@imag.fr)
 */
public interface CacheManager {

    /**
     *    This  method allows a  in-memory object (incarnation)  to be
     * placed in the cache.  Such object is identified by the provided
     * lid.  If  there is currently an object  associated with lid, an
     * Exception is thrown.
     *
     * @param id    The corresponding object identifier.
     *               (It must be generated elsewhere, e.g. by a naming
     *               service).
     * @param object The object to be cached.
     * @return the new CacheEntry add in the cache
     *
     * @exception CacheException If object/lid are already registered
     *            or either object or lid are null.
     **/
    CacheEntry bind(Object id, Object object) throws CacheException;

    /**
     * This method searches the object, identified by lid, within the
     * cache. If such an object has been previously added to the cache
     * the corresponding reference to it will be returned as a result.
     * NULL is returned otherwise. Note that objects must be added to
     * the cache to make it available through it, and that objects could
     * be evicted by the internal replacement algorithm.
     *
     * @param id The corresponding object identifier.
     * @return    The reference to the CacheEntry associated to the id.
     *            NULL otherwise.
     *
     **/
    CacheEntry lookup(Object id);

    /**
     * This method notifies the cache manager the intention to use
     * the object identified by lid. The cache manager will mark
     * internally this object, so that this object would not be
     * evicted from the cache.
     *
     * @param ce
     *
     * @exception CacheException If the object is not in the cache.
     **/
    void fix(CacheEntry ce) throws CacheException;

    /**
     * This method is used to notify the cache manager that the object
     * identified by lid is not longer to be used. After executing this
     * operation, the corresponding object is candidate for eviction,
     * according to the replacement algorithm. That means, that the object
     * will remain in the cache, but it may be evicted.
     *
     * @param ce
     *
     * @exception CacheException If the object has not been previously fixed.
     */
    void unfix(CacheEntry ce) throws CacheException;

    /**
     * Called whenever an object has been accessed. Thus, this method,
     * gives hints about recency/frequency of use. This hints are used
     * within the replacement algorithm.
     *
     * @param entry The cache entry that has been accessed.
     *  @exception CacheException  Whenever an  internal error  occurs.
     **/
    void touch(CacheEntry entry) throws CacheException;
}
