/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.cache.lib;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.LifeCycleController;

import org.objectweb.perseus.cache.api.CacheAttributeController;
import org.objectweb.perseus.cache.api.CacheCapacityEvent;
import org.objectweb.perseus.cache.api.CacheCapacityEventListener;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheEntryFactory;
import org.objectweb.perseus.cache.api.CacheEvent;
import org.objectweb.perseus.cache.api.CacheEventListener;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.perseus.cache.api.FixableCacheEntry;
import org.objectweb.perseus.cache.api.InvalidCacheEntryException;
import org.objectweb.perseus.cache.api.UnbindManager;
import org.objectweb.perseus.cache.api.CacheFullException;
import org.objectweb.perseus.cache.replacement.api.ReplacementManager;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.HashSet;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Collections;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;

/**
 * This class is an implementation of a cache. The data structure is a Hahsed
 * map between cache entry identifier and an entry instance.
 * This class is a fractal component.
 *
 * @author S.Chassande-Barrioz, E.Bruneton
 */
public class BasicCacheManager
		implements
		CacheManager,
		UnbindManager,
		BindingController,
		LifeCycleController,
		CacheAttributeController {

	public final static String DEFAULT_AUTO_CLEAN_SIZE = "7%";
	public final static String CACHE_ENTRY_FACTORY_BINDING = "cache-entry-factory";
	public final static String REPLACEMENT_MANAGER_BINDING = "replacement-manager";
	public final static String CACHE_LISTENER_BINDING = "cache-listeners";
	public final static String CACHE_CAPACTITY_LISTENER_BINDING = "cache-capacity-listeners";


	/**
	 * This class is a basic implementation of the CacheEvent interface.
	 */
	protected class Event implements CacheEvent {

		public int eventId = 0;
		public CacheEntry ce = null;
		public Object oid = null;
		public int cacheSize = -2;

		public Event(int eventid, Object oid, int cachesize) {
			eventId = eventid;
			this.oid = oid;
			cacheSize = cachesize;
		}

		public Event(int eventid, CacheEntry _ce, int cachesize) {
			eventId = eventid;
			ce = _ce;
			cacheSize = cachesize;
		}

		public int getEventId() {
			return eventId;
		}

		public Object getCeIdentifier() {
			return oid;
		}

		public CacheEntry getEntry() {
			return ce;
		}
	}

	/**
	 * This class is a basic implementation of the CacheCapacityEvent interface.
	 */
	protected class CapacityEvent implements CacheCapacityEvent {

		public int eventId = 0;
		public int oldCacheSize = -2;
		public int cacheSize = -2;

		public CapacityEvent(int eventid, int oldCacheSize, int cacheSize) {
			this.eventId = eventid;
			this.cacheSize = cacheSize;
			this.oldCacheSize = oldCacheSize;
		}

		public int getEventId() {
			return eventId;
		}

		public int getOldSize() {
			return oldCacheSize;
		}

		public int getSize() {
			return cacheSize;
		}
	}

	/**
	 * This field is the factory of CacheEntry instances.
	 */
	protected CacheEntryFactory cef = null;

	protected ReplacementManager rm = null;

	/**
	 * This field is the cache. This is a Map which the key are object
	 * identifiers and the values are FixableCacheEntry instances.
	 */
	protected Map cache;

	/**
	 * Queue used to register the weak references stored in the cache.
	 */
	protected ReferenceQueue queue;

	/**
	 * This field is the max size of the cache.
	 */
	protected int cacheMaxSize;

	protected BackgroundCleaner bgcleaner;
	protected int cacheThresHold;
	protected String cacheThresHoldStr;

	protected Logger logger = null;
	protected Logger bgclogger = null;

	/**
	 * This field is the list of CacheEventListener registered.
	 */
	protected HashMap cels;

	/**
	 * This field is the list of CacheEventListener registered.
	 */
	protected HashMap ccels;

	private String autoCleanSizeStr;

	private int autoCleanSize;

	private boolean started = false;

	/**
	 * It builds a BasicCacheManager with an empty CacheEventListener list,
	 * an empty cache and the default max size.
	 */
	public BasicCacheManager() {
		cels = new HashMap();
		ccels = new HashMap();
		cache = new HashMap();
		queue = new ReferenceQueue();
		cacheMaxSize = NO_LIMIT;
		started = false;
		autoCleanSizeStr = null;
		cacheThresHold = 0;
	}

	public BackgroundCleaner getBgcleaner() {
		return bgcleaner;
	}

	//IMPLEMENTATION OF THE LifeCycleController INTERFACE //
	//----------------------------------------------------//

	public String getFcState() {
		return started ? STARTED : STOPPED ;
	}

	public void startFc() {
		if (!started) {
			started = true;
			autoCleanSize = getValue(autoCleanSizeStr);
			cacheThresHold = getValue(cacheThresHoldStr);
		}
	}

	public void stopFc() {
		if (started) {
			started = false;
			try {
				// resizes the cache to 0 in order to clear it.
				setMaxObjects(0);
			} catch (CacheException e) {
			}
		}
	}

	//IMPLEMENTATION OF THE UserBindingControler INTERFACE //
	//-----------------------------------------------------//

    public String[] listFc() {
        HashSet itfs = new HashSet();
        itfs.add(CACHE_ENTRY_FACTORY_BINDING);
        itfs.add(REPLACEMENT_MANAGER_BINDING);
        itfs.add(CACHE_LISTENER_BINDING);
        itfs.add(CACHE_CAPACTITY_LISTENER_BINDING);
        return (String[])itfs.toArray(new String[itfs.size()]);
    }

	public Object lookupFc(String s) {
		if (CACHE_ENTRY_FACTORY_BINDING.equals(s)) {
			return cef;
		} else if (REPLACEMENT_MANAGER_BINDING.equals(s)) {
			return rm;
		} else if (s.startsWith(CACHE_LISTENER_BINDING)) {
			return cels.get(s);
		} else if (s.startsWith(CACHE_CAPACTITY_LISTENER_BINDING)) {
			return ccels.get(s);
		} else {
			return null;
		}
	}

	public void bindFc(String s, Object o) {
		if (CACHE_ENTRY_FACTORY_BINDING.equals(s)) {
			cef = (CacheEntryFactory) o;
		} else if (REPLACEMENT_MANAGER_BINDING.equals(s)) {
			rm = (ReplacementManager) o;
			bgcleaner = new BackgroundCleaner(rm, bgclogger);
		} else if (s.startsWith(CACHE_LISTENER_BINDING)) {
			cels.put(s, o);
		} else if (s.startsWith(CACHE_CAPACTITY_LISTENER_BINDING)) {
			ccels.put(s, o);
		} else if ("monolog-factory".equals(s)) {
			bgclogger = ((LoggerFactory) o)
					.getLogger(logger.getName() + ".bgcleaner");
		} else if ("logger".equals(s)) {
			logger = (Logger) o;
		}
	}

	public void unbindFc(String s) {
		if (CACHE_ENTRY_FACTORY_BINDING.equals(s)) {
			cef = null;
		} else if (REPLACEMENT_MANAGER_BINDING.equals(s)) {
			rm = null;
			if (bgcleaner != null) {
				bgcleaner.stop();
				bgcleaner = null;
			}
		} else if (s.startsWith(CACHE_LISTENER_BINDING)) {
			cels.remove(s);
		} else if (s.startsWith(CACHE_CAPACTITY_LISTENER_BINDING)) {
			ccels.remove(s);
		}
	}

	//IMPLEMENTATION OF THE CacheAttributeControler INTERFACE //
	//--------------------------------------------------------//

	public void setMaxObjects(int size)
			throws IllegalArgumentException, CacheException {
        if (logger != null && logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "resize the cache: "
                    + (size == NO_LIMIT ? "UNLIMITED" : "" + size));
        if (size == NO_LIMIT) {
            cacheMaxSize = NO_LIMIT;
            return;
        }
        if (size < 0)
            throw new IllegalArgumentException(
                    "Impossible to set the cache size to a negative value: " + size);
		synchronized (cache) {
            CacheCapacityEvent ev = new CapacityEvent(
                    CacheCapacityEvent.NOTIFY_CACHE_RESIZE, cache.size(), size);
            if (logger != null && logger.isLoggable(BasicLevel.DEBUG))
                logger.log(BasicLevel.DEBUG,
                    "resize the cache: " + cacheMaxSize + " to " + size
                        + " (Nd elem=" + cache.size() + ")");
			int delta = size - cache.size();
			if (delta < 0) {
				delta = -delta;
				int free = rm.forceFree(delta);
				if (delta > free)
					throw new CacheException("It stays " + (delta - free)
							+ " resources which have not been free");
			}
			cacheMaxSize = size;
			for (Iterator it = ccels.values().iterator(); it.hasNext();) {
				((CacheCapacityEventListener) it.next()).cacheResized(ev);
			}
		}
	}

	public int getMaxObjects() {
		return cacheMaxSize;
	}

	public int getCurrentSize() {
		return cache.size();
	}

	public Collection getCurrentEntryIdentifiers() {
		synchronized(cache) {
			if (cache.size() == 0) {
				return Collections.EMPTY_SET;
			}
			return Collections.unmodifiableCollection(cache.keySet());
		}
	}

	public void setMemorySize(int size) throws IllegalArgumentException {
		if (size < 0)
			throw new IllegalArgumentException(
					"Impossible to set the cache size to a negative value: " + size);
	}

	public int getMemorySize() {
		return 0;
	}

	public int getCurrentMemorySize() {
		return 0;
	}

	public void setAutoCleanSize(String size) {
		autoCleanSizeStr = size;
		autoCleanSize = getValue(autoCleanSizeStr);
	}

	public String getAutoCleanSize() {
		return autoCleanSizeStr;
	}

	public void setAutoCleanThreshold(String size) {
		cacheThresHoldStr = size;
		cacheThresHold = getValue(cacheThresHoldStr);
	}

	public String getAutoCleanThreshold() {
		return cacheThresHoldStr;
	}

	private int getValue(String str) {
		if (cacheMaxSize == NO_LIMIT) {
			return 0;
		}
		if (str == null) {
			str = "75%";
		}
		int i = str.indexOf('%');
		if (i == -1) {
			return Integer.parseInt(str);
		} else {
			int res = Integer.parseInt(str.substring(0, i));
			return (cacheMaxSize * res) / 100;
		}
	}

	//IMPLEMENTATION OF THE CacheManager INTERFACE //
	//---------------------------------------------//

	public CacheEntry bind(Object id, Object object) throws CacheException {
		if (id == null || object == null)
			throw new IllegalArgumentException(
					"Impossible to add a cache entry with a null identifier or object");
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
					"bind an element in the cache, id:" + id + ", object=" + object);
		synchronized (cache) {
			clean();

			// checks that the cache is not full
			int size = cache.size();
			if (cacheMaxSize != NO_LIMIT) {
				if (size >= cacheMaxSize) {
					// removes one element
					rm.forceFree(autoCleanSize);
					size = cache.size();
					if (size >= cacheMaxSize) {
						// if the cache is still full, calls gc and waits some time
						forceClean();
						size = cache.size();
						if (size >= cacheMaxSize) {
							// if the cache is really full, throws an exception
							throw new CacheFullException();
						}
					}
				} else if (cacheThresHold > 0 && bgcleaner != null
						&& (size + 1) >= cacheThresHold) {
					bgcleaner.backgroungCleanup(autoCleanSize);
				}
			}

			// checks that 'id' is not already bound
			WeakCacheEntry ref = (WeakCacheEntry) cache.get(id);
			if (ref != null) {
				FixableCacheEntry entry = (FixableCacheEntry) ref.get();
				//if (entry != null && !(entry.getCeObject() == object && ref.ce == null)) {
				if (entry != null && ref.ce != null) {
					throw new CacheException("Identifier is already bound with "
							+ (entry.getCeObject() == object ? "the same" : "another")
							+ " object: \nid=" + id
                        + "\nobj=" + object
                        + "\nobjInCache=" + entry.getCeObject());
				}
			}

			FixableCacheEntry ce = cef.create(id, object);
			cache.put(id, new WeakCacheEntry(ce, queue));
			rm.addForReplacement(ce);
			CacheEvent ev =
					new Event(CacheEvent.NOTIFY_BIND, ce, ++size);
			logger.log(BasicLevel.DEBUG, "notify the bind");
			for (Iterator it = cels.values().iterator(); it.hasNext();) {
				((CacheEventListener) it.next()).entryBound(ev);
			}
			return ce;
		}
	}

	public CacheEntry lookup(Object id) {
		if (id == null)
			throw new IllegalArgumentException(
					"Impossible to find cache entry with a null identifier");
		synchronized (cache) {
			try {
                CacheEntry ce = get(id);
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "lookup: id=" + id + " // ce =" + ce);
                }
                return ce;
			} catch (InvalidCacheEntryException e) {
				return null;
			}
		}
	}

	public void fix(CacheEntry _ce) throws CacheException {
		if (_ce == null)
			throw new IllegalArgumentException(
					"Impossible to fix a null cache entry");
		synchronized (_ce) {
			FixableCacheEntry ce = get(_ce.getCeIdentifier());
			ce.fixCe();
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG,
						"Fix an element in the cache, id:" + ce.getCeIdentifier()
						+ " / count:" + ce.getCeFixCount());
		}
	}

	public void unfix(CacheEntry ce) throws CacheException {
		if (ce == null)
			throw new IllegalArgumentException(
					"Impossible to unfix a null cache entry");
		synchronized (ce) {
			((FixableCacheEntry) ce).unfixCe();
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG,
						"UnFix an element in the cache, id:" + ce.getCeIdentifier()
						+ " / count:" + ((FixableCacheEntry) ce).getCeFixCount());
		}
	}

	public void touch(CacheEntry ce) throws CacheException {
		if (ce == null)
			throw new IllegalArgumentException(
					"Impossible to touch a null cache entry");
		synchronized (ce) {
			// Check that the entry exists in the cache
			rm.adjustForReplacement((FixableCacheEntry) ce);
		}
	}

	//IMPLEMENTATION OF THE UnbindManager INTERFACE //
	//----------------------------------------------//

	public boolean unbind(Object oid, boolean force) throws CacheException {
		synchronized (cache) { //to prevent lookup, bind in concurrence
			return _unbind(oid, force);
		}
	}
	private boolean _unbind(Object oid, boolean force) throws CacheException {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
			        "Unbind an element from the cache, id:" + oid);
		clean();
		WeakCacheEntry ref = (WeakCacheEntry) cache.get(oid);
		if (ref == null) {
			return true;
		}
		FixableCacheEntry ce = (FixableCacheEntry) ref.get();
		if (ce == null) {
			throw new InvalidCacheEntryException();
		}
		synchronized(ce) {
			if (ce.getCeFixCount() > 0) {
				return false;
			}
			// removes the strong reference to the real entry,
			// but does not remove it from 'cache'
			ref.ce = null;
			if (force) {
				cache.remove(oid);
				CacheEvent ev =
						new Event(CacheEvent.NOTIFY_UNBIND, oid, cache.size());
				for (Iterator it = cels.values().iterator(); it.hasNext();) {
					((CacheEventListener) it.next()).entryUnbound(ev);
				}
				rm.removeForReplacement(oid);
			}
			return force;
		}
	}

	public Collection unbindAll(Collection oids, boolean force) throws CacheException {
		ArrayList unbound = new ArrayList(oids.size());
		synchronized (cache) { //to prevent lookup, bind in concurrence
			if (cache.size() == 0) {
				unbound.addAll(oids);
				return unbound;
			}
			Iterator it = oids.iterator();
			while(it.hasNext()) {
				Object oid = it.next();
				if (!unbound.contains(oid) && _unbind(oid, force)) {
					unbound.add(oid);
				}
			}
			if (!force) {
				/* In case of force == false, we run the GC in order to really
				   evict the entries */
				forceClean();
				clean();
				if (cache.size() == 0) {
					// Optimization when the user has asked an evict all on the
					// cache
					unbound = new ArrayList(oids);
					return unbound;
				}
				// Check if some entries has been removed
				it = oids.iterator();
				while(it.hasNext()) {
					Object oid = it.next();
					WeakCacheEntry ref = (WeakCacheEntry) cache.get(oid);
					if (ref == null) {
						if (!unbound.contains(oid)) {
							unbound.add(oid);
						}
					}
				}
			}
		}
		return unbound;
	}

	public Collection unbindUnfixed(boolean force) throws CacheException {
		return Collections.EMPTY_SET;
	}

	//----------------------------------------------//

	private FixableCacheEntry get(Object id) throws InvalidCacheEntryException {
		Reference ref = (Reference) cache.get(id);
		if (ref == null) {
			/* The cache can be resized during the call, and the entry can
			disappear temporaly. Then to really known if the entry does not
			exist a synchrnonization of the cache is neeeded. */
			synchronized(cache) {
				ref = (Reference) cache.get(id);
			}
			if (ref == null) {
				throw new InvalidCacheEntryException(id);
			}
		}
		FixableCacheEntry entry = (FixableCacheEntry) ref.get();
		if (entry == null) {
			throw new InvalidCacheEntryException();
		}
		//Re-born the entry ;-)
		((WeakCacheEntry) ref).ce = entry;
		return entry;
	}

	/**
	 * Removes all entries that have been garbage collected. Does not block.
	 */
	private void clean() {
		while (true) {
			WeakCacheEntry ref = (WeakCacheEntry) queue.poll();
			if (ref != null && ref.ce == null) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG,
							"Really remove from the cache, id:" + ref.getIdentifier());
				Object oid = ref.getIdentifier();
				cache.remove(oid);
				CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND,
						oid, cache.size());
				for (Iterator it = cels.values().iterator(); it.hasNext();) {
					((CacheEventListener) it.next()).entryUnbound(ev);
				}
				rm.removeForReplacement(oid);
			} else {
				return;
			}
		}
	}

	/**
	 * Calls the garbage collector and waits until an entry can be removed, or a
	 * timeout period (1 second) has expired.
	 */
	private void forceClean() {
		logger.log(BasicLevel.DEBUG, "forceClean()");
		for (int i = 0; i < 6; ++i) {
			System.gc();
			WeakCacheEntry ref = null;
			try {
				ref = (WeakCacheEntry) queue.remove(50);
			} catch (InterruptedException e) {
			}
			if (ref != null) {
				Object oid = ref.getIdentifier();
				cache.remove(oid);
				CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND,
						oid, cache.size());
				for (Iterator it = cels.values().iterator(); it.hasNext();) {
					((CacheEventListener) it.next()).entryUnbound(ev);
				}
				rm.removeForReplacement(oid);
				return;
			}
		}
	}

	/**
	 * The entries stored in 'cache' are weak references to the real entries.
	 * Each weak entry also stores the id of the entry (this id is used in clean
	 * and forceClean to remove the entry from 'cache')
	 */
	private static class WeakCacheEntry extends WeakReference {

		private final Object id;

		// strong reference to the real entry: prevents it from being garbage
		// collected until the unbind method is called.
		public CacheEntry ce;

		public WeakCacheEntry(CacheEntry ce, ReferenceQueue queue) {
			super(ce, queue);
			this.id = ce.getCeIdentifier();
			this.ce = ce;
		}

		public Object getIdentifier() {
			return id;
		}
	}
}
